open Format
open Unix

type config = {
  results_directory: string;
  include_directory: string;
  lustrec: string;
  gcc: string;
}

let empty_unit () = Array.make_matrix 4 4 (-1)

let pp_report_unit fmt =
  Array.iter (Array.iter (fun n ->
      if n = -1 then fprintf fmt "nan " else fprintf fmt "%d " n))

module SMap = Map.Make(String)

let empty_report = SMap.empty

let pp_report fmt =
  fprintf fmt "@[<v>@[<h>\
               #file L-1G0 L-1G1 L-1G2 L-1G3 \
               L1G0 L1G1 L1G2 L1G3 \
               L2G0 L2G1 L2G2 L2G3 \
               L3G0 L3G1 L3G2 L3G3 \
               @]@ %a@]@."
    (fun fmt -> SMap.iter (fun x -> fprintf fmt "@[<h>%s %a@]@ " x pp_report_unit))

let report_file cfg = Filename.concat cfg.results_directory "report"

let write_report cfg report =
  let oc = open_out (report_file cfg) in
  let fmt = formatter_of_out_channel oc in
  pp_report fmt report;
  close_out oc

let config_format_tags =
  set_tags true;
  let mark_open_stag = function
    | String_tag "bold" ->
      "\x1b[1m"
    | String_tag "red" ->
      "\x1b[31m"
    | String_tag "magenta" ->
      "\x1b[35m"
    | String_tag "green" ->
      "\x1b[32m"
    | _ ->
      ""
  in
  let mark_close_stag _ = "\x1b[0m" in
  let stag_fns = get_formatter_stag_functions () in
  set_formatter_stag_functions { stag_fns with mark_open_stag; mark_close_stag }

let pp_header fmt = fprintf fmt "@{<bold>%s:@}"
let err_f = "/tmp/err"

let read_whole_file f =
  let ch = open_in f in
  let s = really_input_string ch (in_channel_length ch) in
  close_in ch;
  s

let lustrec_params cfg o f =
  [
    "-d";
    cfg.results_directory;
    "-O";
    (match o with
     | 0 -> "-1"
     | _ -> string_of_int o);
    f;
  ]

let gcc_params cfg o f =
  [
    "-O" ^ string_of_int o;
    "-fstack-usage";
    "-mfloat-abi=hard";
    "-march=armv7-a";
    "-mfpu=vfpv3-d16";
    "-nostdlib";
    "-specs=nano.specs";
    "-specs=nosys.specs";
    "-I" ^ cfg.include_directory;
    cfg.include_directory ^ "/arrow_noalloc.c";
    "eabi_mock.c";
    f;
    (* "-lgcc" *)
  ]

let stack_usage log =
  let open Re.Str in
  let reg = "[ \t]\\([0-9]+\\)[ \t]" in
  let rec get su n =
    try
      let n = search_forward (regexp reg) log n in
      let su = su + (matched_group 1 log |> int_of_string) in
      get su (n + 1)
    with Not_found -> su
  in
  get 0 0

let pp_warning fmt =
  fprintf fmt "@{<magenta>%s@}"

let pp_error fmt =
  fprintf fmt "@{<red>%s@}"

let warning err_f =
  let log = read_whole_file err_f in
  printf "@ %a" pp_warning log

let error err_f =
  let log = read_whole_file err_f in
  printf "@ %a" pp_error log;
  exit 1

let lustrec_compile cfg o f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.lustrec
      (lustrec_params cfg o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 -> ()
  | _ ->
    error err_f

let gcc_compile cfg report_unit lustrec_o gcc_o f =
  let f' = Filename.remove_extension f ^ ".c" in
  let f_c = Filename.concat cfg.results_directory f' in
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.gcc
      (gcc_params cfg gcc_o f_c)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 ->
    let su_f = "a-" ^ Filename.remove_extension f ^ ".su" in
    let su = read_whole_file su_f in
    report_unit.(lustrec_o).(gcc_o) <- stack_usage su
  | _ -> warning err_f

let compile cfg fs =
  printf "@.%a@." pp_header "Compilation + stack usage estimation tests";
  let n = List.length fs in
  let n_f = float_of_int n in
  List.fold_left (fun (k, report) f ->
      let p = float_of_int k *. 100. /. n_f in
      printf "%3.0f%% @[<v 2>%s" p (Filename.remove_extension f);
      let k = k + 1 in
      let report_unit = empty_unit () in

      (* Compile with LustreC *)
      for lustrec_o = 0 to 3 do
        lustrec_compile cfg lustrec_o f;

        (* Compile with GCC *)
        for gcc_o = 0 to 3 do
          gcc_compile cfg report_unit lustrec_o gcc_o f;
        done;

      done;
      let report =
        SMap.add Filename.(basename (remove_extension f)) report_unit report
      in
      printf "@]@.";
      write_report cfg report;
      k, report)
    (0, empty_report) fs

let expand_path path =
  let home = Unix.getenv "HOME" in
  let open Re.Str in
  global_replace (regexp "~") home path

let set_config () =
  let root = Sys.argv.(1) in
  let lustrec = Sys.argv.(2) in
  let tests_directory = Sys.getcwd () |> Filename.basename in
  let result_dir = root ^ "/../../results/" in
  (try Unix.mkdir result_dir 0o755 with _ -> ());
  let results_directory = result_dir ^ tests_directory in
  (try Unix.mkdir results_directory 0o755 with _ -> ());
  let include_directory = root ^ "/include" in
  let gcc = expand_path "arm-none-eabi-gcc" in
  { results_directory; include_directory; lustrec; gcc; }

let () =
  let cfg = set_config () in
  let lus_fs = Sys.argv.(3) |> Re.Str.(split (regexp_string " ")) in
  let report = compile cfg lus_fs in
  ignore report
