open Format
open Unix

type config = {
  results_directory: string;
  include_directory: string;
  lustrec: string;
  gcc: string;
  compcert: string;
  velus: string;
  otawa: string;
}

type report_unit = {
  lustrec_gcc_wcet: int Array.t Array.t;     (* 4 * 4 *)
  lustrec_compcert_wcet: int Array.t;        (* 4 *)
  velus_wcet: int Array.t                    (* 2 *)
}
let empty_unit () = {
  lustrec_gcc_wcet = Array.make_matrix 4 4 (-1);
  lustrec_compcert_wcet = Array.make 4 (-1);
  velus_wcet = Array.make 2 (-1);
}

let pp_report_unit fmt { lustrec_gcc_wcet; lustrec_compcert_wcet; velus_wcet } =
  let pp n = if n = -1 then fprintf fmt "nan " else fprintf fmt "%d " n in
  Array.iter (Array.iter pp) lustrec_gcc_wcet;
  Array.iter pp lustrec_compcert_wcet;
  Array.iter pp velus_wcet

module SMap = Map.Make(String)

let empty_report = SMap.empty

let pp_report fmt =
  fprintf fmt "@[<v>@[<h>\
               #file L-1G0 L-1G1 L-1G2 L-1G3 \
               L1G0 L1G1 L1G2 L1G3 \
               L2G0 L2G1 L2G2 L2G3 \
               L3G0 L3G1 L3G2 L3G3 \
               L-1C L1C L2C L3C \
               Vnf V@]@ %a@]@."
    (fun fmt -> SMap.iter (fun x -> fprintf fmt "@[<h>%s %a@]@ " x pp_report_unit))

let report_file cfg = Filename.concat cfg.results_directory "report"

let write_report cfg report =
  let oc = open_out (report_file cfg) in
  let fmt = formatter_of_out_channel oc in
  pp_report fmt report;
  close_out oc

let config_format_tags =
  set_tags true;
  let mark_open_stag = function
    | String_tag "bold" ->
      "\x1b[1m"
    | String_tag "red" ->
      "\x1b[31m"
    | String_tag "magenta" ->
      "\x1b[35m"
    | String_tag "green" ->
      "\x1b[32m"
    | _ ->
      ""
  in
  let mark_close_stag _ = "\x1b[0m" in
  let stag_fns = get_formatter_stag_functions () in
  set_formatter_stag_functions { stag_fns with mark_open_stag; mark_close_stag }

let pp_header fmt = fprintf fmt "@{<bold>%s:@}"
let err_f = "/tmp/err"

let read_whole_file f =
  let ch = open_in f in
  let s = really_input_string ch (in_channel_length ch) in
  close_in ch;
  s

let get_step f =
  let open Filename in
  let rec chop f =
    try chop (chop_extension f) with _ -> f
  in
  chop (basename f)

let lustrec_params cfg o f =
  [
    "-velus";
    "-d";
    cfg.results_directory;
    "-O";
    (match o with
     | 0 -> "-1"
     | _ -> string_of_int o);
    f;
  ]

let velus_params o f =
  (match o with
   | 0 -> ["-nofusion"]
   | _ -> [])
  @ [
    "-main";
    get_step f;
    f
  ]

let gcc_params cfg o f =
  [
    "-O" ^ string_of_int o;
    "-mfloat-abi=hard";
    "-march=armv7-a";
    "-mfpu=vfpv3-d16";
    "-nostdlib";
    "-specs=nano.specs";
    "-specs=nosys.specs";
    "-I" ^ cfg.include_directory;
    cfg.include_directory ^ "/arrow_noalloc.c";
    "eabi_mock.c";
    f;
    (* "-lgcc" *)
  ]

let compcert_params cfg f =
  [
    "-nostdlib";
    "-fstruct-passing";
    "-I" ^ cfg.include_directory;
    cfg.include_directory ^ "/arrow_noalloc.c";
    "eabi_mock.c";
    f;
  ]

let otawa_params step = [
  "-s";
  "trivial";
  "a.out";
  step
]

let cycles log =
  let open Re.Str in
  let reg = "\\([0-9]+\\) cycles" in
  search_forward (regexp reg) log 0 |> ignore;
  matched_group 1 log |> int_of_string

let has_velus_error log =
  let open Re.Str in
  let reg =
    "Velus output Error:.+"
  in
  try
    search_forward (regexp reg) log 0 |> ignore;
    Some (matched_string log)
  with Not_found -> None

let pp_warning fmt =
  fprintf fmt "@{<magenta>%s@}"

let pp_error fmt =
  fprintf fmt "@{<red>%s@}"

let warning err_f =
  let log = read_whole_file err_f in
  printf "@ %a" pp_warning log

let lustrec_compile cfg o f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.lustrec
      (lustrec_params cfg o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 -> ()
  | _ ->
    let log = read_whole_file err_f in
    match has_velus_error log with
    | Some m ->
      printf "@ %a" pp_warning m;
      raise Exit
    | None -> printf "@ %a" pp_error log

let otawa_wcet cfg step =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.otawa
      (otawa_params step)
  in
  let ic = open_process_in cmd in
  let log = try input_line ic with _ -> "" in
  match close_process_in ic with
  | WEXITED n when n <> 2 ->
    Some (cycles log)
  | _ ->
    warning err_f;
    None

let gcc_compile cfg report_unit lustrec_o gcc_o f =
  let f' = Filename.remove_extension f ^ ".c" in
  let f_c = Filename.concat cfg.results_directory f' in
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.gcc
      (gcc_params cfg gcc_o f_c)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 ->
    let step = get_step f ^ "_step" in
    begin match otawa_wcet cfg step with
      | Some wcet ->
        report_unit.lustrec_gcc_wcet.(lustrec_o).(gcc_o) <- wcet
      | None -> ()
    end
  | _ -> warning err_f

let compcert_compile' cfg f_c step update =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.compcert
      (compcert_params cfg f_c)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 ->
    begin match otawa_wcet cfg step with
      | Some wcet ->
        update wcet
      | None -> ()
    end
  | _ -> warning err_f

let compcert_compile cfg report_unit lustrec_o f =
  let f' = Filename.remove_extension f ^ ".c" in
  let f_c = Filename.concat cfg.results_directory f' in
  let step = get_step f ^ "_step" in
  compcert_compile' cfg f_c step (fun wcet ->
      report_unit.lustrec_compcert_wcet.(lustrec_o) <- wcet)

let velus_compile cfg report_unit velus_o f =
  let f' = Filename.remove_extension f ^ ".velus.lus" in
  let f = Filename.concat cfg.results_directory f' in
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.velus
      (velus_params velus_o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 ->
    let f_s = Filename.remove_extension f ^ ".s" in
    let step = "fun$step$" ^ get_step f in
    compcert_compile' cfg f_s step (fun wcet ->
        report_unit.velus_wcet.(velus_o) <- wcet)
  | _ -> warning err_f

let compile cfg fs =
  printf "@.%a@." pp_header "Compilation + WCET estimation tests";
  let n = List.length fs in
  let n_f = float_of_int n in
  List.fold_left (fun (k, report) f ->
      let p = float_of_int k *. 100. /. n_f in
      printf "%3.0f%% @[<v 2>%s" p (Filename.remove_extension f);
      let k = k + 1 in
      let report_unit = empty_unit () in

      let report = try
          (* Compile with LustreC *)
          for lustrec_o = 0 to 3 do
            lustrec_compile cfg lustrec_o f;

            (* Compile with GCC *)
            for gcc_o = 0 to 3 do
              gcc_compile cfg report_unit lustrec_o gcc_o f;
            done;

            (* Compile with CompCert *)
            compcert_compile cfg report_unit lustrec_o f;

          done;

          (* Compile with Velus *)
          for velus_o = 0 to 1 do
            velus_compile cfg report_unit velus_o f;
          done;
          SMap.add Filename.(basename (remove_extension f)) report_unit report
        with Exit -> report
      in
      printf "@]@.";
      write_report cfg report;
      k, report)
    (0, empty_report) fs

let expand_path path =
  let home = Unix.getenv "HOME" in
  let open Re.Str in
  global_replace (regexp "~") home path

let set_config () =
  let root = Sys.argv.(1) in
  let lustrec = Sys.argv.(2) in
  let tests_directory = Sys.getcwd () |> Filename.basename in
  let result_dir = root ^ "/../../results/" in
  (try Unix.mkdir result_dir 0o755 with _ -> ());
  let results_directory = result_dir ^ tests_directory in
  (try Unix.mkdir results_directory 0o755 with _ -> ());
  let include_directory = root ^ "/include" in
  let gcc = expand_path "arm-none-eabi-gcc" in
  let otawa = expand_path "owcet" in
  let velus = expand_path "~/velus/velus" in
  let compcert = Filename.(concat (dirname velus) "CompCert/ccomp") in
  { results_directory; include_directory; lustrec; gcc; compcert; velus; otawa; }

let () =
  let cfg = set_config () in
  let lus_fs = Sys.argv.(3) |> Re.Str.(split (regexp_string " ")) in
  let report = compile cfg lus_fs in
  ignore report
